// MotionCapture.cpp : Motion Capture with Kinect and Direct3D
// This source code is licensed under the MIT license. Please see the License in License.txt.
//

#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#include <tchar.h>
#include <time.h>
#include <sstream>
#include <exception>

#include <d3d9.h>
#include <d3dx9.h>

#include <objbase.h>
#include <NuiApi.h>

#pragma comment( lib, "d3d9.lib" )
#pragma comment( lib, "d3dx9.lib" )


// EBhẼTCY
static const int WINDOW_WIDTH  = 800;
static const int WINDOW_HEIGHT = 600;

// EBhẼnh
static HWND g_mainWindowHandle = 0;

// Win32APIɊւ肪ƂɓO
class win32_exception : public std::exception
{
public:
	win32_exception( const char *const msg ) : std::exception( msg ) {}
};

// KinectɊւ肪ƂɓO
class kinect_exception : public std::exception
{
public:
	kinect_exception( const char *const msg ) : std::exception( msg ) {}
};

// Direct3DɊւ肪ƂɓO
class d3d_exception : public std::exception
{
public:
	d3d_exception( const char *const msg ) : std::exception( msg ) {}
};

/*----- t[[gv -----*/

static clock_t g_fpsClk = clock();
static int g_fps = 0;
static int g_fpsCur = 0;

// t[[gvD1`悷閈1x̊֐ĂяoD1bƂɍXVC[͎ľܓ
int updateFps()
{
	const clock_t current = clock();
	const int diff = current - g_fpsClk;
	if( diff >= CLOCKS_PER_SEC ) {
		g_fps = static_cast<int>( ( static_cast<float>( g_fpsCur ) / ( diff / CLOCKS_PER_SEC ) + 0.5f ) );
		g_fpsCur = 0;
		g_fpsClk = ( current / CLOCKS_PER_SEC ) * CLOCKS_PER_SEC + diff % CLOCKS_PER_SEC;
	}
	g_fpsCur++;
	return g_fps;
}

/* ----- Kinect ----- */

// Kinect̃CX^X
static INuiSensor* g_sensor = nullptr;

// RGB摜̃Xg[ƁC擾Ɠ邽߂̃nh
static HANDLE g_rgbStream = INVALID_HANDLE_VALUE;
static HANDLE g_rgbHandle = INVALID_HANDLE_VALUE;

// Skeleton̎擾Ɠ邽߂̃nh
static HANDLE g_skeleHandle = INVALID_HANDLE_VALUE;

// Skeletoñt[
static NUI_SKELETON_FRAME g_skeleFrame;

// Kinect̃`g[^[̊px
static float g_sensorTiltAngle = 0.0f;
static float g_sensorTiltAnglePrev[ 10 ] = { 0.0f };

// Kinect̃\[XJ
void releaseKinect()
{
	if( g_sensor ) {
		g_sensor->NuiSkeletonTrackingDisable();
		g_sensor->NuiShutdown();
	}
	CloseHandle( g_rgbHandle );
	CloseHandle( g_skeleHandle );
}

// Kinect
void initKinect()
{
	HRESULT hResult;

	// Kinect
	hResult = NuiCreateSensorByIndex( 0, &g_sensor );
	if( FAILED( hResult ) ) {
		throw kinect_exception(
			"Error : NuiCreateSensorByIndex\nKinectłĂȂCgpł" );
	}

	hResult = g_sensor->NuiInitialize( NUI_INITIALIZE_FLAG_USES_COLOR | NUI_INITIALIZE_FLAG_USES_SKELETON );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiInitialize" );
	}

	// KinectƓ邽߂̃Cxg쐬

	g_skeleHandle = CreateEvent( nullptr, TRUE, FALSE, nullptr );
	if( FAILED( g_skeleHandle ) ) {
		throw win32_exception( "Error : CreateEvent" );
	}

	g_rgbHandle = CreateEvent( nullptr, TRUE, FALSE, nullptr );
	if( FAILED( g_rgbHandle ) ) {
		throw win32_exception( "Error : CreateEvent" );
	}

	// RGB摜擾邽߂̃Xg[J
	hResult = g_sensor->NuiImageStreamOpen( NUI_IMAGE_TYPE_COLOR,
		NUI_IMAGE_RESOLUTION_640x480, 0, 2, g_rgbHandle, &g_rgbStream );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiImageStreamOpen" );
	}

	// Skeleton̒ǐՂL
	hResult = g_sensor->NuiSkeletonTrackingEnable( g_skeleHandle, 0 );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiSkeletonTrackingEnable" );
	}

	// `g[^[̊px擾
	long tiltAngle;
	hResult = g_sensor->NuiCameraElevationGetAngle( &tiltAngle );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiCameraElevationGetAngle" );
	}
	for( int i = 0; i < ARRAYSIZE( g_sensorTiltAnglePrev ); i++ ) {
		g_sensorTiltAnglePrev[ i ] = static_cast<float>( tiltAngle );
	}

}

/* ----- Direct3D ----- */

// Bone̒[cm]
static const float BONE_LENGTH[ 20 ] = {
	0.0f,  // Root
	5.1f,  // K -> w
	28.3f, // w -> 
	21.5f, //  -> 
	19.8f, //  -> 
	24.3f, //  -> I
	26.5f, // I -> 
	8.2f,  //  -> 
	19.8f, // î͍EΏ̂Ȃ̂œlƂj
	24.3f,
	26.5f,
	8.2f,
	10.0f, // K -> 
	35.8f, //  -> G
	35.2f, // G -> 
	11.5f, //  -> 
	10.0f,
	35.8f,
	35.2f,
	11.5f
};

// Bonẽ[gnʂ܂ł̋[cm]
static const float BONE_ROOT_DISTANCE = 108.4f;

// Bone̕\ɕKvȌvZʂi[邽߂̍\
struct SkeleState
{
	// Joinẗʒu
	D3DXVECTOR3 endJointPos[ NUI_SKELETON_POSITION_COUNT ];

	// [Jϊs
	D3DXMATRIX localTrans[ NUI_SKELETON_POSITION_COUNT ];

	// gbLO
	bool boneTracked[ NUI_SKELETON_POSITION_COUNT ];

	// ]ʒuȂǂLۂ
	bool isValid;

	// o[ϐSď
	void Reset()
	{
		for( int i = 0; i < NUI_SKELETON_POSITION_COUNT; i++ ) {
			endJointPos[ i ].x = 0.0f;
			endJointPos[ i ].y = 0.0f;
			endJointPos[ i ].z = 0.0f;
			D3DXMatrixIdentity( &localTrans[ i ] );
			boneTracked[ i ] = false;
			isValid = false;
		}
	}
};
static SkeleState g_skeleState[ NUI_SKELETON_COUNT ];

// D3DIuWFNg
static IDirect3D9 *g_d3d = nullptr;

// D3DfoCX
static IDirect3DDevice9 *g_d3ddev = nullptr;

// Kinect擾RGB摜\邽߂̃eNX`[
static IDirect3DTexture9 *g_kinectRgbTex = nullptr;

// ʂƎlp\3Df̒_obt@[
static IDirect3DVertexBuffer9 *g_planeVB = nullptr;
static IDirect3DVertexBuffer9 *g_pyramidVB = nullptr;

// 
static ID3DXFont *g_font = nullptr;

// TCY
static const int DEBUG_FONT_SIZE = 24;

// WƃeNX`[WFVF
struct FvfTex {
	FLOAT x, y, z;
	FLOAT u, v;
};

// WƐFFVF
struct FvfDiffuse {
	FLOAT x, y, z;
	DWORD color;
};

// Direct3D
void releaseD3D()
{
	if( g_d3d )           g_d3d->Release();           g_d3d = nullptr;
	if( g_d3ddev )        g_d3ddev->Release();        g_d3ddev = nullptr;
	if( g_kinectRgbTex )  g_kinectRgbTex->Release();  g_kinectRgbTex = nullptr;
	if( g_planeVB )       g_planeVB->Release();       g_planeVB = nullptr;
	if( g_pyramidVB )     g_pyramidVB->Release();     g_pyramidVB = nullptr;
	if( g_font )          g_font->Release();          g_font = nullptr;
}

// Direct3D
void initD3D()
{
	HRESULT hResult;

	// Direct3DIuWFNg쐬
	g_d3d = Direct3DCreate9( D3D_SDK_VERSION );
	if( g_d3d == nullptr )
		throw d3d_exception( "Error : Direct3DCreate9" );

	// Direct3DfoCX쐬
	D3DPRESENT_PARAMETERS presentParam;
	ZeroMemory( &presentParam, sizeof presentParam );
	presentParam.SwapEffect = D3DSWAPEFFECT_DISCARD;
	presentParam.EnableAutoDepthStencil = TRUE;
	presentParam.AutoDepthStencilFormat = D3DFMT_D24X8;
	presentParam.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	presentParam.hDeviceWindow = g_mainWindowHandle;
	presentParam.Windowed = TRUE;

	hResult = g_d3d->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
		g_mainWindowHandle, D3DCREATE_HARDWARE_VERTEXPROCESSING,
		&presentParam, &g_d3ddev );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3D9#CreateDevice" );
	}

	// RGB摜\邽߂̃eNX`[쐬
	hResult = g_d3ddev->CreateTexture( 640, 480, 1,
		D3DUSAGE_DYNAMIC, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT,
		&g_kinectRgbTex, nullptr );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DDevice9#CreateTexture" );
	}

	// ʂ̃|S쐬
	void *lockedMem;
	static const FvfTex planeVertex[] = {
		{ -1.0f,  1.0f, 0.0f, 0.0f, 0.0f },
		{  1.0f,  1.0f, 0.0f, 1.0f, 0.0f },
		{ -1.0f, -1.0f, 0.0f, 0.0f, 1.0f },
		{ -1.0f, -1.0f, 0.0f, 0.0f, 1.0f },
		{  1.0f,  1.0f, 0.0f, 1.0f, 0.0f },
		{  1.0f, -1.0f, 0.0f, 1.0f, 1.0f },
	};
	hResult = g_d3ddev->CreateVertexBuffer( sizeof( planeVertex ), 0,
		D3DFVF_XYZ | D3DFVF_TEX1, D3DPOOL_MANAGED, &g_planeVB, nullptr );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DDevice9#CreateVertexBuffer" );
	}
	
	hResult = g_planeVB->Lock( 0, 0, &lockedMem, D3DLOCK_DISCARD );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DVertexBuffer9#Lock" );
	}

	memcpy( lockedMem, planeVertex, sizeof( planeVertex ) );

	hResult = g_planeVB->Unlock();
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DVertexBuffer9#Unock" );
	}

	// lp̃|S쐬
	static FvfDiffuse pyramidVertex[] = {
		{ -3.0f,  0.0f, -3.0f, 0xFF33AAAA },
		{  3.0f,  0.0f, -3.0f, 0xFF33AAAA },
		{ -3.0f,  0.0f,  3.0f, 0xFF33AAAA },
		
		{ -3.0f,  0.0f,  3.0f, 0xFF33AAAA },
		{  3.0f,  0.0f, -3.0f, 0xFF33AAAA },
		{  3.0f,  0.0f,  3.0f, 0xFF33AAAA },

		{  3.0f,  0.0f, -3.0f, 0xFFEE33BB },
		{ -3.0f,  0.0f, -3.0f, 0xFFEE33BB },
		{  0.0f,  1.0f,  0.0f, 0xFFEE33BB },
		
		{ -3.0f,  0.0f, -3.0f, 0xFFCC33BB },
		{ -3.0f,  0.0f,  3.0f, 0xFFCC33BB },
		{  0.0f,  1.0f,  0.0f, 0xFFCC33BB },
		
		{ -3.0f,  0.0f,  3.0f, 0xFFAA33BB },
		{  3.0f,  0.0f,  3.0f, 0xFFAA33BB },
		{  0.0f,  1.0f,  0.0f, 0xFFAA33BB },
		
		{  3.0f,  0.0f,  3.0f, 0xFF8833BB },
		{  3.0f,  0.0f, -3.0f, 0xFF8833BB },
		{  0.0f,  1.0f,  0.0f, 0xFF8833BB },
	};

	hResult = g_d3ddev->CreateVertexBuffer( sizeof( pyramidVertex ), 0,
		D3DFVF_XYZ | D3DFVF_DIFFUSE, D3DPOOL_MANAGED, &g_pyramidVB, nullptr );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DDevice9#CreateVertexBuffer" );
	}
	
	hResult = g_pyramidVB->Lock( 0, 0, &lockedMem, D3DLOCK_DISCARD );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DVertexBuffer9#Lock" );
	}

	memcpy( lockedMem, pyramidVertex, sizeof( pyramidVertex ) );

	hResult = g_pyramidVB->Unlock();
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DVertexBuffer9#Unock" );
	}

	// `̏
	D3DXFONT_DESC fontDesc;
	ZeroMemory( &fontDesc, sizeof( fontDesc ) );
	fontDesc.Width = DEBUG_FONT_SIZE / 2;
	fontDesc.Height = DEBUG_FONT_SIZE;
	fontDesc.Weight = FW_NORMAL;
	fontDesc.MipLevels = 1;
	fontDesc.Quality = DEFAULT_QUALITY;
	fontDesc.CharSet = DEFAULT_CHARSET;
	_tcscpy_s( fontDesc.FaceName, _T( "Meiryo UI" ) );

	hResult = D3DXCreateFontIndirect( g_d3ddev, &fontDesc, &g_font );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : D3DXCreateFontIndirect" );
	}
}

/*----- 烁C̏ -----*/

// Kinect擾
// XVꂽƂtrueԂ
bool capture()
{
	HRESULT hResult;
	DWORD dResult;

	// Kinect̃f[^̎擾Ă邩ǂׂ
	const HANDLE events[] = { g_rgbHandle, g_skeleHandle };
	dResult = WaitForMultipleObjects( ARRAYSIZE( events ), events, true, 0 );
	// ܂̂Ƃ͎擾Ȃ
	if( dResult == WAIT_TIMEOUT ) {
		return false;
	}
	if( dResult == 0xFFFFFFFF ) {
		throw win32_exception( "Error : MsgWaitForMultipleObjects" );
	}

	// RGB摜̐Vt[擾
	NUI_IMAGE_FRAME rgbFrame;
	INuiFrameTexture* rgbTexture;
	NUI_LOCKED_RECT rgbKinectRect;

	hResult = g_sensor->NuiImageStreamGetNextFrame( g_rgbStream, 0, &rgbFrame );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiImageStreamGetNextFrame" );
	}

	// Rs[RGB摜̓ǂݎJn
	rgbTexture = rgbFrame.pFrameTexture;
	hResult = rgbTexture->LockRect( 0, &rgbKinectRect, nullptr, 0 );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : INuiFrameTexture#LockRect" );
	}
	
	// Rs[̃eNX`[݂̏Jn
	D3DLOCKED_RECT rgbD3DRect;
	hResult = g_kinectRgbTex->LockRect( 0, &rgbD3DRect, nullptr, D3DLOCK_DISCARD );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DTexture9#LockRect" );
	}

	// 1sRs[JԂ
	for( int row = 0; row < ( rgbKinectRect.size / rgbKinectRect.Pitch ); row++ ) {
		// Rs[̃|C^
		const byte *psrc = rgbKinectRect.pBits + row * rgbKinectRect.Pitch;
		// Rs[̃|C^
		byte *pdest = reinterpret_cast< byte* >( rgbD3DRect.pBits ) + row * rgbD3DRect.Pitch;
		memcpy( pdest, psrc, rgbKinectRect.Pitch );
	}

	// eNX`[RGB摜̓]
	hResult = g_kinectRgbTex->UnlockRect( 0 );
	if( FAILED( hResult ) ) {
		throw d3d_exception( "Error : Error : IDirect3DTexture9#UnlockRect" );
	}

	hResult = rgbTexture->UnlockRect( 0 );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : INuiFrameTexture#UnlockRect" );
	}

	// RGB摜̃t[
	hResult = g_sensor->NuiImageStreamReleaseFrame( g_rgbStream, &rgbFrame );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiImageStreamReleaseFrame" );
	}

	// Skeleton̐Vt[擾
	hResult = g_sensor->NuiSkeletonGetNextFrame( 0, &g_skeleFrame );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiSkeletonGetNextFrame" );
	}

	// SkeletoñX[WO
	const NUI_TRANSFORM_SMOOTH_PARAMETERS customSmooth =
		{0.5f, 0.5f, 0.5f, 0.05f, 0.04f}; // ftHg
		//{0.5f, 0.1f, 0.5f, 0.1f, 0.1f}; // 蕽
		//{0.0f, 1.0f, 0.0f, 0.001f, 0.001f}; // Ȃ
	hResult = g_sensor->NuiTransformSmooth( &g_skeleFrame, &customSmooth );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiTransformSmooth" );
	}

	// `g[^[̊px擾
	long tiltAngle;
	hResult = g_sensor->NuiCameraElevationGetAngle( &tiltAngle );
	if( FAILED( hResult ) ) {
		throw kinect_exception( "Error : NuiCameraElevationGetAngle" );
	}

	// pxXV
	for( int i = ARRAYSIZE( g_sensorTiltAnglePrev ) - 1; i >= 1; i-- ) {
		g_sensorTiltAnglePrev[ i ] = g_sensorTiltAnglePrev[ i - 1 ];
	}
	g_sensorTiltAnglePrev[ 0 ] = static_cast<float>( tiltAngle );

	// px̕ϒl
	float angle = 0.0f;
	for( int i = 0; i < ARRAYSIZE( g_sensorTiltAnglePrev ); i++ ) {
		angle += g_sensorTiltAnglePrev[ i ];
	}

	g_sensorTiltAngle = angle / ARRAYSIZE( g_sensorTiltAnglePrev );

	// CxgVOiԂɖ߂
	ResetEvent( g_rgbHandle );
	ResetEvent( g_skeleHandle );

	return true;
}

// Matrix4D3DXMATRIXɕϊ
static void convertMat4ToD3DXMat( const Matrix4 &src, D3DXMATRIX &dest )
{
	dest._11 = src.M11; dest._12 = src.M12; dest._13 = src.M13; dest._14 = src.M14;
	dest._21 = src.M21; dest._22 = src.M22; dest._23 = src.M23; dest._24 = src.M24;
	dest._31 = src.M31; dest._32 = src.M32; dest._33 = src.M33; dest._34 = src.M34;
	dest._41 = src.M41; dest._42 = src.M42; dest._43 = src.M43; dest._44 = src.M44;
}

// WϊƃgbLOԂSkeleStateɊi[
static void calcSkeleState(
	const NUI_SKELETON_DATA *skele,
	NUI_SKELETON_POSITION_INDEX start,
	NUI_SKELETON_POSITION_INDEX end,
	const D3DXVECTOR3 &jointStartPos,
	const Matrix4 &rotate,
	float boneLength,
	SkeleState *state
	)
{
	D3DXMATRIX mat;
	D3DXMATRIX matScale, matRotate, matTrans;
	const D3DXVECTOR3 yVec( 0.0f, 1.0f, 0.0f ); // y1LтĂxNg
	D3DXVECTOR3 ret;

	// 3DfYɊg傷
	D3DXMatrixScaling( &matScale, 1.0f, boneLength, 1.0f );

	// ]
	convertMat4ToD3DXMat( rotate, matRotate );
	
	// ړ
	D3DXMatrixTranslation( &matTrans, jointStartPos.x, jointStartPos.y, jointStartPos.z );

	// ϊs쐬
	D3DXMatrixMultiply( &mat, &matScale, &matRotate );
	D3DXMatrixMultiply( &state->localTrans[ end ], &mat, &matTrans );

	// _Wϊ
	D3DXVec3TransformCoord( &state->endJointPos[ end ], &yVec, &state->localTrans[ end ] );

	// BonẽgbLOԂݒ肷
	if( skele->eSkeletonPositionTrackingState[ start ] == NUI_SKELETON_POSITION_TRACKED
		&& skele->eSkeletonPositionTrackingState[ end ] == NUI_SKELETON_POSITION_TRACKED ) {
		state->boneTracked[ end ] = true;
	}
	else {
		state->boneTracked[ end ] = false;
	}
}

// Bone Orientation\ɕKvȏ߂SkeleStateɊi[
static void setSkeleStateFromOrient(
	const NUI_SKELETON_DATA *skele,
	const NUI_SKELETON_BONE_ORIENTATION *orient,
	SkeleState *skeleState )
{
	NUI_SKELETON_POSITION_INDEX jointStart, jointEnd;

	jointStart = NUI_SKELETON_POSITION_HIP_CENTER;
	jointEnd   = NUI_SKELETON_POSITION_HIP_CENTER;

	// Root͕\ł{[Ȃ̂ŁA[JϊšvZ͏ȗ

	// gbLOԂ擾
	skeleState->boneTracked[ jointEnd ] =
		skele->eSkeletonPositionTrackingState[ jointEnd ] == NUI_SKELETON_POSITION_TRACKED;

	// RootȊOBoneɑ΂calcSkeleState()Ăяo
	for( int i = 1; i < NUI_SKELETON_POSITION_COUNT; i++ )
	{
		jointStart = orient[ i ].startJoint;
		jointEnd   = orient[ i ].endJoint;

		calcSkeleState( skele, jointStart, jointEnd, skeleState->endJointPos[ jointStart ],
			orient[ jointEnd ].absoluteRotation.rotationMatrix, BONE_LENGTH[ jointEnd ], skeleState );
	}
}

// Kinectf[^擾A]s̐Ȃǂ̌vZOɏĂ
void preprocess()
{
	HRESULT hResult;
	for( int i = 0; i < NUI_SKELETON_COUNT; i++ ) {
		NUI_SKELETON_DATA *skele = &g_skeleFrame.SkeletonData[ i ];
		g_skeleState[ i ].Reset();

		// ǐՉ\ȏԂɂ
		if( skele->eTrackingState == NUI_SKELETON_TRACKED ) {
			// ]߁AKvȌvZs
			NUI_SKELETON_BONE_ORIENTATION orient[ NUI_SKELETON_POSITION_COUNT ];
			hResult = NuiSkeletonCalculateBoneOrientations( skele, orient );

			// ]vZłƂ
			if( hResult == S_OK ) {
				setSkeleStateFromOrient( skele, orient, &g_skeleState[ i ] );
				g_skeleState[ i ].isValid = true;
			}
		}
	}
}

// Skeleton`悷
static void drawSkeleton( NUI_SKELETON_DATA *skele, float floorHeight, float tiltAngle, SkeleState *skeleState )
{
	// `g[^[̉]pxSkeletoñ[g̈ʒuɁA[hWnɔzu邽߂̕ϊs
	D3DXMATRIX matWorld, matTiltRot, matSkeleTrans;

	D3DXMatrixRotationX( &matTiltRot, D3DXToRadian( -1.0f * tiltAngle ) );

	D3DXMatrixTranslation(
		&matSkeleTrans,
		skele->SkeletonPositions[ NUI_SKELETON_POSITION_HIP_CENTER ].x * 100.0f,
		( skele->SkeletonPositions[ NUI_SKELETON_POSITION_HIP_CENTER ].y + floorHeight ) * 100.0f - BONE_ROOT_DISTANCE,
		( skele->SkeletonPositions[ NUI_SKELETON_POSITION_HIP_CENTER ].z ) * 100.0f );

	D3DXMatrixMultiply( &matWorld, &matSkeleTrans, &matTiltRot );

	// 3Df`悷
	// Root͕`łȂ̂Ŕ΂
	for( int i = 1; i < NUI_SKELETON_POSITION_COUNT; i++ ) {
		D3DXMATRIX matLocalWorld;

		// ŏIIȕϊs߂
		D3DXMatrixMultiply( &matLocalWorld, &skeleState->localTrans[ i ], &matWorld );

		// [hϊsݒ肷
		g_d3ddev->SetTransform( D3DTS_WORLD, &matLocalWorld );
		
		// BonegbLOłĂƂ͕ʂɁAłĂȂƂ̓C[t[ŕ`悷
		if( skeleState->boneTracked[ i ] ) {
			g_d3ddev->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			g_d3ddev->DrawPrimitive( D3DPT_TRIANGLELIST, 0, 6 );
		}
		else {
			g_d3ddev->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
			g_d3ddev->DrawPrimitive( D3DPT_TRIANGLELIST, 0, 6 );
		}
	}

	// `@ftHgɖ߂
	g_d3ddev->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
}

// `悷
void draw()
{
	HRESULT hResult;

	D3DXVECTOR3 vecEye, vecAt, vecUp;
	D3DXMATRIX matWorld, matView, matProj;
	D3DVIEWPORT9 viewport = { 0, 0, WINDOW_WIDTH, WINDOW_HEIGHT, 0.0f, 1.0f };

	RECT textRect = { 0, DEBUG_FONT_SIZE, WINDOW_WIDTH, WINDOW_HEIGHT };


	// foCXXgւ̑Ή
	hResult = g_d3ddev->TestCooperativeLevel();
	if( hResult == D3DERR_DEVICELOST ) {
		return;
	}
	else if( hResult == D3DERR_DEVICENOTRESET ) {
		releaseD3D();
		initD3D();
	}

	// obt@[NA
	g_d3ddev->Clear( 0, nullptr, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0xFF332211, 1.0f, 0 );
	
	// `Jn
	g_d3ddev->BeginScene();
	
	// gȂ
	g_d3ddev->SetRenderState( D3DRS_LIGHTING, FALSE );

	// C[t[\̂߁CJOgȂ
	g_d3ddev->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

	// [xobt@[g
	g_d3ddev->SetRenderState( D3DRS_ZENABLE, TRUE );

	// _̐ݒ
	vecEye = D3DXVECTOR3( 0.0f, 0.0f,   0.0f );
	vecAt  = D3DXVECTOR3( 0.0f, 0.0f, 400.0f );
	vecUp  = D3DXVECTOR3( 0.0f, 1.0f,   0.0f );
	D3DXMatrixLookAtLH( &matView, &vecEye, &vecAt, &vecUp );
	g_d3ddev->SetTransform( D3DTS_VIEW, &matView );

	// ϊ
	D3DXMatrixPerspectiveFovLH(
		&matProj, D3DXToRadian( NUI_CAMERA_DEPTH_NOMINAL_VERTICAL_FOV + 5.0f ),
		static_cast<float>( WINDOW_WIDTH ) / WINDOW_HEIGHT, 1.0f, 1000.f );
	g_d3ddev->SetTransform( D3DTS_PROJECTION, &matProj );

	// r[|[g̐ݒ
	g_d3ddev->SetViewport( &viewport );

	/* Skeleton̕\ */

	// _̐ݒ
	g_d3ddev->SetFVF( D3DFVF_XYZ | D3DFVF_DIFFUSE );
	g_d3ddev->SetStreamSource( 0, g_pyramidVB, 0, sizeof( FvfDiffuse ) );
	
	// eNX`[̐ݒ
	g_d3ddev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_DISABLE );
	g_d3ddev->SetTexture( 0, nullptr );

	// ̋擾
	float floorHeight;
	if( g_skeleFrame.vFloorClipPlane.y > FLT_EPSILON ) {
		floorHeight = g_skeleFrame.vFloorClipPlane.w;
	}
	// 擾łȂƂ́Cli1[gjg
	else {
		floorHeight = 1.0f;
	}

	// Skeleton`悷
	int numTrackedSkele = 0;
	for( int i = 0; i < NUI_SKELETON_COUNT; i++ ) {
		NUI_SKELETON_DATA skele = g_skeleFrame.SkeletonData[ i ];

		// ǐՉ\ȏԂɂ
		if( skele.eTrackingState == NUI_SKELETON_TRACKED ) {
			numTrackedSkele++;

			// ]vZłƂ
			if( g_skeleState[ i ].isValid ) {
				drawSkeleton( &skele, floorHeight, g_sensorTiltAngle, &g_skeleState[ i ] );
			}
			else {
				g_font->DrawText( nullptr, _T( "]vZłȂSkeleton܂" ), -1, &textRect, 0, 0xFFFFFFAA );
				textRect.top += DEBUG_FONT_SIZE;
			}
		}
	}

	// ̋\
	if( g_skeleFrame.vFloorClipPlane.w > FLT_EPSILON ) {
		std::stringstream bufss;
		const Vector4 f = g_skeleFrame.vFloorClipPlane;
		bufss << "Kinect̏̋ : " << f.w << "[m]";
		g_font->DrawTextA( nullptr, bufss.str().c_str(), -1, &textRect, 0, 0xFFFFFFFF );
		textRect.top += DEBUG_FONT_SIZE;
	}
	// ołȂƂA̎|\
	else {
		g_font->DrawText( nullptr,
			_T( "oł܂DJ̈ʒuςĂ" ), -1, &textRect, 0, 0xFFFFFFAA );
		textRect.top += DEBUG_FONT_SIZE;
	}

	// `悷SkeletonȂƂA̎|\
	if( numTrackedSkele == 0 ) {
		g_font->DrawText( nullptr, _T( "Skeletonoł܂" ), -1, &textRect, 0, 0xFFFFFFAA );
		textRect.top += DEBUG_FONT_SIZE;
	}
	// Ƃ́A̐\
	else {
		std::stringstream bufss;
		bufss << "Skeleton̐ : " << numTrackedSkele;
		g_font->DrawTextA( nullptr, bufss.str().c_str(), -1, &textRect, 0, 0xFFAAFFFF );
		textRect.top += DEBUG_FONT_SIZE;
	}

	/* RGB摜̕\ */
	
	// [xobt@[gȂ
	g_d3ddev->SetRenderState( D3DRS_ZENABLE, FALSE );

	// _̐ݒ
	g_d3ddev->SetFVF( D3DFVF_XYZ | D3DFVF_TEX1 );
	g_d3ddev->SetStreamSource( 0, g_planeVB, 0, sizeof( FvfTex ) );

	// eNX`[̐ݒ
	g_d3ddev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	g_d3ddev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
	g_d3ddev->SetTexture( 0, g_kinectRgbTex );

	// Wϊ̐ݒiϊsȂj
	D3DXMatrixIdentity( &matWorld );
	D3DXMatrixIdentity( &matView );
	D3DXMatrixIdentity( &matProj );
	g_d3ddev->SetTransform( D3DTS_WORLD, &matWorld );
	g_d3ddev->SetTransform( D3DTS_VIEW, &matView );
	g_d3ddev->SetTransform( D3DTS_PROJECTION, &matProj );

	// r[|[g̐ݒ
	viewport.X      = WINDOW_WIDTH * 3 / 4;
	viewport.Width  = WINDOW_WIDTH * 1 / 4;
	viewport.Y      = 0;
	viewport.Height = WINDOW_HEIGHT * 1 / 4;
	g_d3ddev->SetViewport( &viewport );
	
	// `
	g_d3ddev->DrawPrimitive( D3DPT_TRIANGLELIST, 0, 2 );

	/* t[[g̕\ */

	// r[|[g̐ݒ
	viewport.X      = 0;
	viewport.Width  = WINDOW_WIDTH;
	viewport.Y      = 0;
	viewport.Height = WINDOW_HEIGHT;
	g_d3ddev->SetViewport( &viewport );

	// t[[g̎擾ƕ`
	std::stringstream ssFps;
	int fps = updateFps();
	ssFps << "FPS : " << fps;
	textRect.top = 0;
	g_font->DrawTextA( nullptr, ssFps.str().c_str(), -1, &textRect, 0, 0xFFFFFFFF );

	/* \ */

	// `I
	g_d3ddev->EndScene();

	// fBXvCɓ]
	hResult = g_d3ddev->Present( nullptr, nullptr, nullptr, nullptr );
	if( hResult != D3DERR_DEVICELOST && FAILED( hResult ) ) {
		throw d3d_exception( "Error : IDirect3DDevice9#Present" );
	}
}

// EBhEvV[W
LRESULT CALLBACK WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
	PAINTSTRUCT ps;
	HDC hdc;

	switch( message ) {
	case WM_KEYDOWN:
		// OpenCVhighgui̋ɍ킹
		if( wParam == VK_ESCAPE ) {
			PostMessage( hWnd, WM_DESTROY, 0, 0 );
			return 0;
		}
		break;

	case WM_PAINT:
		hdc = BeginPaint( hWnd, &ps );
		EndPaint( hWnd, &ps );
		break;

	case WM_DESTROY:
		PostQuitMessage( 0 );
		break;

	default:
		return DefWindowProc( hWnd, message, wParam, lParam );
	}
	return 0;
}

// EBhE쐬
static HWND setupWindow( int width, int height )
{
	WNDCLASSEX wcex;
	wcex.cbSize         = sizeof( WNDCLASSEX );
	wcex.style          = CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc    = WndProc;
	wcex.cbClsExtra     = 0;
	wcex.cbWndExtra     = 0;
	wcex.hInstance      = (HMODULE)GetModuleHandle( 0 );
	wcex.hIcon          = nullptr;
	wcex.hCursor        = LoadCursor( nullptr, IDC_ARROW );
	wcex.hbrBackground  = (HBRUSH)(COLOR_WINDOW + 1);
	wcex.lpszMenuName   = nullptr;
	wcex.lpszClassName  = _T( "KinectWindowClass" );
	wcex.hIconSm        = nullptr;
	if( !RegisterClassEx( &wcex ) ) {
		throw win32_exception( "Error : RegisterClassEx" );
	}

	RECT rect = { 0, 0, width, height };
	AdjustWindowRect( &rect, WS_OVERLAPPEDWINDOW, FALSE );
	const int windowWidth  = ( rect.right  - rect.left );
	const int windowHeight = ( rect.bottom - rect.top );

	HWND hWnd = CreateWindow( _T("KinectWindowClass"), _T("# Kinect Skeleton Sample #"),
		WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, windowWidth, windowHeight,
		nullptr, nullptr, nullptr, nullptr );
	if( !hWnd ) {
		throw win32_exception( "Error : CreateWindow" );
	}

	return hWnd;
}

// Gg[|Cg
int WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, int )
{
	MSG msg;
	ZeroMemory( &msg, sizeof msg );

	// Debugł͗OJ󂯎CReleaseł̓bZ[W\ċIɏI
#ifdef NDEBUG
	try { 
#endif
	
	g_mainWindowHandle = setupWindow( WINDOW_WIDTH, WINDOW_HEIGHT );
	ShowWindow( g_mainWindowHandle, SW_SHOW );
	UpdateWindow( g_mainWindowHandle );

	// KinectDirect3D̏
	initKinect();
	initD3D();

	while( msg.message != WM_QUIT ) {
		BOOL r = PeekMessage( &msg, nullptr, 0, 0, PM_REMOVE );
		if( r == 0 ) {
			bool updated = capture();
			if( updated ) {
				preprocess();
			}
			draw();
		}
		else {
			DispatchMessage( &msg );
		}
	}

#ifdef NDEBUG
	} catch( std::exception &e ) {
		MessageBoxA( g_mainWindowHandle, e.what(), "Kinect Sample - 肪܂", MB_ICONSTOP ); }
#endif

	// KinectDirect3D̏I
	releaseKinect();
	releaseD3D();

	return static_cast<int>( msg.wParam );
}

